/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/***************************************************************************
 *            conv_opt.cc
 *
 *  Thu Jan 24 12:26:30 CET 2008
 *  Copyright 2008 Jonas Suhr Christensen and Bent Bisballe Nyeng
 *  suhr@daimi.au.dk and deva@daimi.au.dk
 ****************************************************************************/

/*
 *    This is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    It is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this product; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
#include <stdlib.h>
#include <fftw3.h>
#include <sndfile.hh>
#include <string.h>
#include <assert.h>
#include "sys/timeb.h"
#include <math.h>
#include <cmath>
#include <errno.h>

#define BUFFER_SIZE_MS 1000 //ms 113,  // Find a size with small prime factors

#define MEASURE

#ifdef FFTW_FLOAT
#define FFTW_DATATYPE fftwf_complex
#define FFTW_DATATYPE_REAL float 
#define FFTW_PLAN_DFT fftwf_plan_dft_1d
#define FFTW_PLAN_DFT_REAL fftwf_plan_dft_r2c_1d
#define FFTW_PLAN_DFT_REAL_INVERSE fftwf_plan_dft_c2r_1d
#define FFTW_EXECUTE fftwf_execute
#define FFTW_PLAN fftwf_plan
#define FFTW_DESTROY_PLAN fftwf_destroy_plan
#define FFTW_FREE fftwf_free
#define FFTW_MALLOC fftwf_malloc
#define FFTW_INIT_THREADS fftwf_init_threads
#define FFTW_PLAN_WITH_NTHREADS fftwf_plan_with_nthreads
#define FFTW_CLEANUP_THREADS fftwf_cleanup_threads
typedef float sample_t;
#else
#define FFTW_DATATYPE fftw_complex
#define FFTW_DATATYPE_REAL double 
#define FFTW_PLAN_DFT fftw_plan_dft_1d
#define FFTW_PLAN_DFT_REAL fftw_plan_dft_r2c_1d
#define FFTW_PLAN_DFT_REAL_INVERSE fftw_plan_dft_c2r_1d
#define FFTW_EXECUTE fftw_execute
#define FFTW_PLAN fftw_plan
#define FFTW_DESTROY_PLAN fftw_destroy_plan
#define FFTW_FREE fftw_free
#define FFTW_MALLOC fftw_malloc
#define FFTW_INIT_THREADS fftw_init_threads
#define FFTW_PLAN_WITH_NTHREADS fftw_plan_with_nthreads
#define FFTW_CLEANUP_THREADS fftw_cleanup_threads
typedef double sample_t;
#endif

unsigned int BUFFER_SIZE_SAMPLES;
unsigned long int totaltime;
long int iterations;

/**
 ** Optimizations
 **/
//#define USE_REAL  // Use the real function of the fftw3 library

typedef struct {
  sample_t *data;
  size_t size;
} samples_t;


samples_t *new_samples(sample_t *data, size_t size)
{
	samples_t *samples = (samples_t*)malloc(sizeof(samples_t));
	samples->data = data;
	samples->size = size; 
	return samples;
}

samples_t *read_data(const char *filename, double scale)
{
//	printf("Loading %s ... ", filename); fflush(stdout);

	SndfileHandle sfh(filename);
	size_t size = sfh.seek(0, SEEK_END);
	sfh.seek(0, SEEK_SET);
	sample_t *data = (sample_t*)malloc(sizeof(sample_t) * size);

	sfh.read(data, size); 

//	printf("[%d samples] ... ", size); fflush(stdout);

	for(int i = 0; i < size; i++) {
		data[i] = scale * data[i];
	}
	
  //	for(int i = size / 2; i < (size / 2) +20; i++) printf("\t[%d] %f\n", i, data[i]);

	samples_t *samples = new_samples(data, size);
//	printf("done\n");

	return samples;
}


void write_data(const char *filename, samples_t *samples)
{
//	printf("Saving %s ...", filename); fflush(stdout);

	//	for(int i = samples->size / 2; i < (samples->size / 2) +20; i++)printf("\t[%d] %f\n", i, samples->data[i]);

	SndfileHandle sfh(filename, SFM_WRITE, SF_FORMAT_WAV | SF_FORMAT_PCM_16, 1, 44100);
	
	sfh.write(samples->data, samples->size);

//	printf("done\n");
}

#define RE 0
#define IM 1
#ifdef USE_REAL
static FFTW_DATATYPE *hfft = NULL;
static FFTW_DATATYPE *outm = NULL;
static FFTW_DATATYPE_REAL *out = NULL;
static FFTW_DATATYPE_REAL *inx = NULL;
static FFTW_DATATYPE *xfft = NULL;
#else
static FFTW_DATATYPE *hfft = NULL;
static FFTW_DATATYPE *outm = NULL;
static FFTW_DATATYPE *out = NULL;
static FFTW_DATATYPE *inx = NULL;
static FFTW_DATATYPE *xfft = NULL;
#endif /*USE_REAL*/

void init_fftw(samples_t *h) 
{
#ifdef FFTW_THREADS
  FFTW_INIT_THREADS();
  FFTW_PLAN_WITH_NTHREADS(3);
#endif

  FFTW_PLAN p;

  int size = BUFFER_SIZE_SAMPLES + h->size - 1;
  hfft = (FFTW_DATATYPE*) FFTW_MALLOC(sizeof(FFTW_DATATYPE) * size);

  FFTW_DATATYPE *inh = (FFTW_DATATYPE*) FFTW_MALLOC(sizeof(FFTW_DATATYPE) * size);

  for(int i = 0; i < size; i++) {
    inh[i][RE] = inh[i][IM] = 0;
  }
  for(int i = 0; i < h->size; i++) {
    inh[i][RE] = h->data[i];
  }

  p = FFTW_PLAN_DFT(size, inh, hfft, FFTW_FORWARD, FFTW_ESTIMATE);
  FFTW_EXECUTE(p);
  FFTW_DESTROY_PLAN(p);

  FFTW_FREE(inh); 

  outm = (FFTW_DATATYPE*) FFTW_MALLOC(sizeof(FFTW_DATATYPE) * size);
#ifdef USE_REAL
  out = (FFTW_DATATYPE_REAL*) malloc(sizeof(FFTW_DATATYPE_REAL) * size);
  inx = (FFTW_DATATYPE_REAL*) malloc(sizeof(FFTW_DATATYPE_REAL) * size);
  for(int i = 0; i < size; i++) {
    inx[i] = 0;
  }
#else
  out = (FFTW_DATATYPE*) FFTW_MALLOC(sizeof(FFTW_DATATYPE) * size);
  inx = (FFTW_DATATYPE*) FFTW_MALLOC(sizeof(FFTW_DATATYPE) * size);
  for(int i = 0; i < size; i++) {
    inx[i][RE] = inx[i][IM] = 0;
  }
#endif /*USE_REAL*/
  xfft = (FFTW_DATATYPE*) FFTW_MALLOC(sizeof(FFTW_DATATYPE) * size);


}

void deinit_fftw()
{
  FFTW_FREE(hfft);
  FFTW_FREE(out); 
  FFTW_FREE(outm); 
  FFTW_FREE(inx); 
  FFTW_FREE(xfft);
#ifdef FFTW_THREADS
  FFTW_CLEANUP_THREADS();
#endif
}

#ifdef MEASURE
long static time_used;
#endif
samples_t *conv_fftwc_preprocess(samples_t *x, size_t hsize/*samples_t *h*/)
{
  assert(hfft); // init_fftw not called!
  assert(out); // init_fftw not called!
  assert(outm); // init_fftw not called!
  assert(inx); // init_fftw not called!
  assert(xfft); // init_fftw not called!

#ifdef MEASURE
  struct timeb systime;
  long begin = 0;
  long end = 0;
  ftime(&systime);
  begin = systime.time * 1000 + systime.millitm;
#endif

  static FFTW_PLAN px, pi;
  static bool first = true;

  int size = BUFFER_SIZE_SAMPLES/*x->size*/ + hsize - 1;
  //  printf("N = %d\n", size);

  sample_t *data = (sample_t*)calloc(sizeof(sample_t), size);
  samples_t *y = new_samples(data, size);

#ifdef USE_REAL
   for(int i = 0; i < BUFFER_SIZE_SAMPLES/*x->size*/; i++) {
    inx[i] =  x->data[i];
  }
#else
  for(int i = 0; i < BUFFER_SIZE_SAMPLES/*x->size*/; i++) {
    inx[i][RE] =  x->data[i];
  }
#endif /*USE_REAL*/

  #ifdef USE_REAL
    if(first) {
      px = FFTW_PLAN_DFT_REAL(size, /*x->data*/inx, xfft, FFTW_MEASURE);
    }
  #else
    if(first) {
      px = FFTW_PLAN_DFT(size, inx, xfft, FFTW_FORWARD, FFTW_MEASURE);
    }
  #endif
  FFTW_EXECUTE(px);


  for(int i = 0; i < size; i++ ) {
    outm[i][RE] = hfft[i][RE] * xfft[i][RE] - hfft[i][IM] * xfft[i][IM];
    outm[i][IM] = xfft[i][IM] * hfft[i][RE] + xfft[i][RE] * hfft[i][IM];
  }


  #ifdef USE_REAL
    if(first) {
      pi = FFTW_PLAN_DFT_REAL_INVERSE(size, outm, out, FFTW_MEASURE);
      first = false;
    }
  #else
    if(first) {
      pi = FFTW_PLAN_DFT(size, outm, out, FFTW_BACKWARD, FFTW_MEASURE);
      first = false;
    }
  #endif /*USE_REAL*/
  FFTW_EXECUTE(pi);


#ifdef USE_REAL
  for(int i = 0; i < size; i++) {
//    printf("\tout[%d] = %d\n", i, out[i]);
    data[i] = out[i] / size;
  }
#else
  for(int i = 0; i < size; i++) {
    data[i] = out[i][RE] / size;
  }
#endif /*USE_REAL*/

#ifdef MEASURE
  ftime(&systime);
  end = systime.time * 1000 + systime.millitm;
//  printf("\t Time consumed: %d\n", end - begin);
  time_used = end - begin;
  if(iterations != -1)
    totaltime += end - begin;
  iterations++;
#endif


  return y;
}

sample_t *filter(const sample_t *input, size_t filter_size)
{
	unsigned int buffer_size =  BUFFER_SIZE_SAMPLES + filter_size - 1;
	static sample_t *buffer = (sample_t*)calloc(buffer_size, sizeof(sample_t));
	static unsigned int p = 0;

	samples_t *input_s = new_samples((sample_t *)input, BUFFER_SIZE_SAMPLES);

	samples_t *y;

	y = conv_fftwc_preprocess(input_s, filter_size);

	for(int i = 0; i < BUFFER_SIZE_SAMPLES; i++) {
		buffer[(i + p + filter_size - 1) % buffer_size] = 0; //y->data[i];
	}

	for(int i = 0/* BUFFER_SIZE_SAMPLES*/; i < buffer_size; i++) {
		buffer[(i + p) % buffer_size] += y->data[i];
	}
	
	// alloker returbuffer
	sample_t *output_buffer = (sample_t*)malloc(BUFFER_SIZE_SAMPLES * sizeof(sample_t));

	// kopier retsultat til output buffer
	for(int i = 0; i < BUFFER_SIZE_SAMPLES; i++) {
		output_buffer[i] = buffer[(i + p) % buffer_size];
	}

	p += BUFFER_SIZE_SAMPLES;

	// frigr y
  free(y->data);
  free(y);
	free(input_s);
  
	return output_buffer;
}


samples_t *conv_simple(samples_t *x, samples_t *h)
{
  int size = x->size + h->size - 1;
  sample_t *data = (sample_t*)calloc(sizeof(sample_t), size);
  
  samples_t *y = new_samples(data, size);
  
  
  for(int xn = 0; xn < x->size; xn++) {
    for(int hn = 0; hn < h->size; hn++) {
      y->data[xn + hn] += h->data[hn] * x->data[xn];
    }
  } 
    
  return y;
}

sample_t puls = 0.5;
int main(int argc, char *argv[])
{
	//	y(n)=\sum_{k=-\infty}^\infty h(k)x(n-k)
	
	if(argc != 5) {
		printf("Usage %s [filterfile] [soundfile] [outputfile] [buffer size]\n", argv[0]);
		exit(1);
	}

	int xsz, hsz;

	const char *filterfile = argv[1];
	const char *inputfile = argv[2];
	const char *outputfile = argv[3];
  BUFFER_SIZE_SAMPLES = atoi(argv[4]);

  totaltime = 0;
  iterations = -1;

	samples_t *h = read_data(filterfile, 0.08);
	//	samples_t *h = new_samples(&puls, 0.5);
	samples_t *x = read_data(inputfile, 0.5);
	//	samples_t *x = new_samples(&puls, 1);


	size_t s = h->size + x->size - 1;
	sample_t *d = (sample_t*)malloc(s  * sizeof(sample_t));

	samples_t *y = new_samples(d, s);

  
  // 0 padding
  sample_t *pd = (sample_t*)malloc((y->size) * sizeof(sample_t));
  for(int i = 0; i < x->size; i++) {
    pd[i] = x->data[i];
  }
  for(int i = x->size; i < y->size + (y->size % BUFFER_SIZE_SAMPLES); i++) {
    pd[i] = 0;
  }
  x = new_samples(pd, x->size);

	init_fftw(h);

  int e = 0;

  for(int i = 0; i < y->size; i += BUFFER_SIZE_SAMPLES) {
#ifdef MEASURE
//    printf("[c:%d, s:%d](%d, %d)", i, y->size, BUFFER_SIZE_SAMPLES, e); fflush(stdout);
#endif
    sample_t *data = filter(&(x->data[i]), h->size /*- e*/);

		for(int j = 0; j < BUFFER_SIZE_SAMPLES; j++) {
			y->data[j + i] = data[j]; 
		}

    free(data);
	}

  printf("%d %f\n", h->size, (double)totaltime/(double)iterations);

	deinit_fftw();

	write_data(outputfile, y);

	free(y);

	free(h);
	free(x);


}
